// Port
var port = 49111;

var WebSocket = require(__dirname + "/node_modules/ws").Server;
var config = require(__dirname + "/config.js");

var ws = new WebSocket({port: port});
var clients = [];
var currentID = 0;
var lastDrawingIndex = 0;
var lostDrawings = [];
var lostImages = [];
var mysqlEnabled = false;
var version = "1.3.1";

if(config.authRequired) {
	// Mysql connection
	var mysql = require('mysql');
	var connection =  mysql.createConnection({host: config.host, user: config.user, password: config.password, database: config.database});

	connection.connect(function(err) {
	  if (err) {
		console.log(m.e + "Cannot reach database!");
		config.authRequired = false;
		return;
	  }

        mysqlEnabled = true;
        console.log(m.i + "Connection to database succesful (id: " + connection.threadId + ")");
	});
}

// Colors enum
var c = {
    r: "\033[31m",
    g: "\033[32m",
    y: "\033[33m",
    b: "\033[36m",
    w: "\033[37m"
};

// Messages enum
var m = {
    i: c.g + "[INFO] " + c.w,
    e: c.r + "[ERROR] " + c.w,
    w: c.b + "[WARNING] " + c.w
};

console.log('\033[2J' + m.i + "Server started (listening on port " + port + ")");

var tools = new function(){
	this.broadcast = function(clients, type, data, excludedClient) {
					for(var i = 0; i < clients.length; i++) {
						if(clients[i].data.isAuthenticated || typeof clients[i].data.isAuthenticated  === "undefined")
							if(!excludedClient || excludedClient.data.id != clients[i].data.id) {
								sendMessage(clients[i], type, data);
							}
					}
				};

this.sendMessage = function(client, type, data) {
					try {
						client.send(JSON.stringify({type:type, data:data}));
					} catch(e) {}
				};
}

// Handle new users
ws.on("connection", function(client) {
    console.log(m.i + "New connection (auth required ? " + config.authRequired + ")");
	tools.sendMessage(client, "auth-required", config.authRequired);

	client.data = {};
	client.data.isAuthenticated = false;


	client.on("message", function(message) {
		try {
			message = JSON.parse(message);
		} catch(e) {
			console.warn(m.r + "Cannot parse message: " + message);
		}

		// Server core
		if(client.data.isAuthenticated) {
			switch(message.type) {
				case "pong":
					clearTimeout(client.pingTimeout);
					client.pingTimeout = false;
					break;
				case "mouse-ping":
					message.data.clientId = client.data.id;
					tools.broadcast(clients, "mouse-ping", message.data);
					break;
				case "client-leave":
					disconnect(client);
					break;
				case "chat-message":
                    console.log(m.i + client.data.username + ": " + message.data);
					tools.broadcast(clients, "chat-message", {username:client.data.username, message:message.data});
					break;
				case "change-tool":
                    console.log(m.i + client.data.username + " changed tool to " + message.data.tool);
					client.data.tools = message.data;
					tools.broadcast(clients, "change-tool", {tools: message.data, clientId: client.data.id});
					break;
				case "load-image":
                    /*console.log(m.i + client.data.username + " loaded " + message.data);
					tools.broadcast(clients, "load-image", message.data);*/
					if(!isURL(message.data.url))
						return;
					console.log("Sending image")
					tools.broadcast(clients, "load-image", {clientId: client.data.id, url: message.data.url, x: message.data.x, y: message.data.y, width: message.data.width, height: message.data.height, ldi: lastDrawingIndex});
					client.data.images.push({url: message.data.url, x: message.data.x, y: message.data.y, width: message.data.width, height: message.data.height, ldi: lastDrawingIndex});
					lastDrawingIndex++;
					break;
				case "clear-whiteboard":
                    console.log(m.w + client.data.username + " cleared the whiteboard");
					for(var i = 0; i < clients.length; i++) {
						clients[i].data.drawings = {squares:[], circles:[], triangles:[], points:[], lines: [], images: [], text:[]};
						clients[i].data.undoneDrawings = [];
					}
					lostDrawings = [];
					lostImages = [];
					tools.broadcast(clients, "clear-whiteboard", {clientId: client.data.id});
					break;
				case "draw":
                    console.log(m.i + client.data.username + " drew a " + message.data.type);
					switch(message.data.type) {
						case "point":
							if(message.data.points.length == 0)
								return;
							console.log("Added a point");
							client.data.drawings.points.push({points: message.data.points, size: client.data.tools.size, color: client.data.tools.color, opacity: client.data.tools.opacity, fill: client.data.tools.fill, ldi: lastDrawingIndex});
							break;
						case "square":
							client.data.drawings.squares.push({x: message.data.x, y: message.data.y, width: message.data.width, height: message.data.height, size: client.data.tools.size, color: client.data.tools.color, opacity: client.data.tools.opacity, fill: client.data.tools.fill, ldi: lastDrawingIndex});
							break;
						case "circle":
							client.data.drawings.circles.push({x: message.data.x, y: message.data.y, radius: message.data.radius, color: client.data.tools.color, opacity: client.data.tools.opacity, fill: client.data.tools.fill, size: client.data.tools.size, ldi: lastDrawingIndex});
							break;
						case "triangle":
							client.data.drawings.triangles.push({x: message.data.x, y: message.data.y, mouseX: message.data.mouseX, mouseY: message.data.mouseY, color: client.data.tools.color, opacity: client.data.tools.opacity, fill: client.data.tools.fill, size: client.data.tools.size, ldi: lastDrawingIndex});
							break;
						case "line":
							client.data.drawings.lines.push({x: message.data.x, y: message.data.y, mouseX: message.data.mouseX, mouseY: message.data.mouseY, color: client.data.tools.color, opacity: client.data.tools.opacity, fill: client.data.tools.fill, size: client.data.tools.size, ldi: lastDrawingIndex});
							break;
						/*case "image":
							if(!isURL(message.data.url)) {
								return;
							}
							tools.broadcast(clients, "loading-image", {clientId: client.data.id, url: message.data.url});
							client.data.drawings.images.push({url: message.data.url, x: message.data.x, y: message.data.y, width: message.data.width, height: message.data.height, ldi: lastDrawingIndex});
							break;*/
						case "text":
							client.data.drawings.text.push({text: message.data.text, x: message.data.x, y: message.data.y, color: client.data.tools.color, size: client.data.tools.size,opacity: client.data.tools.opacity, fill: client.data.tools.fill, ldi: lastDrawingIndex});
							break;
					}
					client.data.undoneDrawings = [];
					tools.broadcast(clients, "update-drawings", {clientId: client.data.id, drawings: client.data.drawings});
					lastDrawingIndex++;
					break;
				case "undo":
					var lastAction = {key: 0, type: "", ldi: 0};
					for(key in client.data.drawings) {
						for(var i = 0; i < client.data.drawings[key].length; i++) {
							if(lastAction.ldi <= client.data.drawings[key][i].ldi) {
								lastAction.key = i;
								lastAction.ldi = client.data.drawings[key][i].ldi;
								lastAction.type = key;
							}
						}
					}

					if(typeof(client.data.drawings[lastAction.type]) !== 'undefined') {
						client.data.undoneDrawings.push({type: lastAction.type, drawing: client.data.drawings[lastAction.type][lastAction.key]});
						client.data.drawings[lastAction.type].splice(lastAction.key, 1);
						tools.broadcast(clients, "update-drawings", {clientId: client.data.id, drawings: client.data.drawings});
					}
					break;
				case "redo":
					var lastDrawing = client.data.undoneDrawings[client.data.undoneDrawings.length - 1];
					if(typeof(lastDrawing) === 'undefined')
						return;

					client.data.drawings[lastDrawing.type].push(lastDrawing.drawing);
					client.data.undoneDrawings.splice(client.data.undoneDrawings.length - 1, 1);

					tools.broadcast(clients, "update-drawings", {clientId: client.data.id, drawings: client.data.drawings});
					break;
                case "add-user":
                    if(!checkAdmin(client))
                        break;
						if(!message.data.name && !message.data.pass)
							break;
						connection.query( 'INSERT INTO user VALUES(\'\', '+ connection.escape(message.data.name) +', '+ connection.escape(message.data.pass) +', 0)', function(err) {
                            console.log(m.i + "New user created: " + message.data.name);
							if(err)
								tools.sendMessage(client, "server-message", "Impossible d'ajouter l'utilisateur");
							else
								tools.sendMessage(client, "server-message", "Utilisateur créé!");
						});
                    break;
                case "set-pass":
                    if(!checkAdmin(client))
                        break;
						if(!message.data.name && !message.data.pass)
							break;
						connection.query( 'UPDATE user SET password = ' + connection.escape(message.data.pass) + ' WHERE username = ' + connection.escape(message.data.name), function(err, results) {
                            console.log(m.i + "Password change for " + message.data.name);
							if(!err & results.affectedRows == 0)
								tools.sendMessage(client, "server-message", "Impossible de modifier le mot de passe");
							else
								tools.sendMessage(client, "server-message", "Mot de passe modifié!");
						});
                    break;
                case "remove-user":
                    if(!checkAdmin(client))
                        break;
						connection.query( 'DELETE FROM user WHERE username = ' + connection.escape(message.data.name), function(err, results) {
                            console.log(m.i + message.data.name + " deleted");
							if(!err & results.affectedRows == 0)
								tools.sendMessage(client, "server-message", "Impossible de supprimer l'utilisateur");
							else
								tools.sendMessage(client, "server-message", "Utilisateur supprimé!");

						});
                    break;
                case "kick-user":
                    if(checkAdmin(client))
                        break;
                    var kickOccured = false;
                    for(key in clients) {
                        if(clients[key].data.username == message.data.name) {
                            disconnect(clients[key]);
                            kickOccured = true;
                            break;
                        }
                    }
                    if(kickOccured)
                        tools.broadcast(clients, "server-message", message.data.name + " a été expulsé!");
                    else
                        tools.sendMessage(client, "server-message", "Aucun utilisateur n'utilise ce nom...");
                    break;
                case "toggle-auth":
                    if(!checkAdmin(client))
                        break;
                    config.authRequired = !config.authRequired;
                    if(config.authRequired) {
                        tools.broadcast(clients, "server-message", "Connexion au chat requise pour parler");
                    }
                    else
                        tools.broadcast(clients, "server-message", "Connexion au chat non-requise pour parler.");

                    if(!message.data.removeAll && config.authRequired) {
                        for(key in clients) {
                            if(!clients[key].data.isLogged) {
                                tools.broadcast(clients, "server-message", clients[key].data.username + " a été expulsé!");
                                disconnect(clients[key]);
                            }
                        }
                    }
                    break;
			}
		}
		else {
			switch(message.type) {
				case "auth-try":
					// if config.authrequired
					if(config.authRequired) {
						connection.query( 'SELECT username, password, admin FROM user WHERE username = ' + connection.escape(message.data.username), function(err, rows) {
							if(rows.length == 1 && rows[0].password == message.data.password) {
								createClient(client, message, rows);
							}
							else {
                                console.log(m.e + "Login attempt fail with username: " + message.data.username);
								tools.sendMessage(client, "auth-status", false);
							}
						});
					}
					else {
						createClient(client, message);
					}
				break;
			}
		}
	});
});

function checkAdmin(client) {
    if(client.data.isAdmin) {
        if(!mysqlEnabled) {
            tools.sendMessage(client, "server-message", "Non connecté à la base de données.");
            return false;
        }
        return true;
    }
    else
        tools.sendMessage(client, "server-message", "Permission refusée.");
}

setInterval(function() {
	var d = new Date();
	for(var i = 0; i < clients.length; i++) {
		if(clients[i].pingTimeout)
			continue;

		tools.sendMessage(clients[i], "ping");

		(function(client, i){ clients[i].pingTimeout = setTimeout(function() {
			console.log(m.w + (client.data.username || client.data.id) + "] timed out" );
			disconnect(client);
		}, 5000);})(clients[i], i);
	}
}, 5000);

function createClient(client, message, rows) {
	for(var i = 0; i < clients.length; i++) {
		if(clients[i].data.username == message.data.username) {
			tools.sendMessage(client, "auth-status", false);
			return;
		}
	}

	client.data.tools = {
		size: 12,
		tool: 0,
		color: "#000000",
		opacity: 1,
		fill: false
	};

	client.data.drawings = {squares: [], circles: [], triangles: [], points: [], lines: [], text:[]};
	client.data.images = [];
	client.data.undoneDrawings = [];
	client.data.id = currentID++;
	client.data.username = message.data.username;
	client.data.isAuthenticated = true;

	if(config.authRequired) {
		client.data.isAdmin = rows[0].admin;
        client.data.isLogged = true;
    }
    else {
		client.data.isAdmin = false;
        client.data.isLogged = false;
    }

	clients.push(client);

	var dataToSend = [];
	for(var i = 0; i < clients.length; i++) {
		dataToSend.push(clients[i].data);
	}

	tools.broadcast(clients, "client-join", client.data);
	tools.sendMessage(client, "auth-status", true);

	tools.sendMessage(client, "client-list", dataToSend);
	tools.sendMessage(client, "update-drawings", {clientId: -1, drawings: lostDrawings});
	tools.sendMessage(client, "update-drawings", {clientId: -2, drawings: lostImages});
	tools.sendMessage(client, "own-id", client.data.id);
	console.log(m.i + client.data.username + " connected successfully");
}

function disconnect(client) {
	var tempClient = client;
	if(tempClient.data.drawings.length  != 0)
		lostDrawings.push(tempClient.data.drawings);
	for(var i = 0; i < tempClient.data.images.length; i++)
		lostImages.push(tempClient.data.images[i]);
	console.log(m.i + (client.data.username || client.data.id) + " left");
    client.close();
	clients.splice(clients.indexOf(client), 1);
	if(lostDrawings.length  != 0)
		tools.broadcast(clients, "update-drawings", {clientId: -1, drawings: lostDrawings});
	if(lostImages.length != 0)
		tools.broadcast(clients, "update-drawings", {clientId: -2, drawings: lostImages});
	tools.broadcast(clients, "client-leave", {clientId: tempClient.data.id, queuePosition: tempClient.data.queuePosition});
}

function isURL(str) {
  var pattern = new RegExp('^(https?:\\/\\/)?'+ // protocol
  '((([a-z\\d]([a-z\\d-]*[a-z\\d])*)\\.?)+[a-z]{2,}|'+ // domain name
  '((\\d{1,3}\\.){3}\\d{1,3}))'+ // OR ip (v4) address
  '(\\:\\d+)?(\\/[-a-z\\d%_.~+]*)*'+ // port and path
  '(\\?[;&a-z\\d%_.~+=-]*)?'+ // query string
  '(\\#[-a-z\\d_]*)?$','i'); // fragment locator
  return pattern.test(str);
}
